@@ -13,7 +13,7 @@
 # limitations under the License.
 # ==============================================================================
 
-from typing import List
+from typing import List, Callable
 import atexit
 import contextlib
 import ctypes
@@ -44,19 +44,24 @@ class BlueFogBasics(object):
         self._is_topo_weighted = False
         self.warn_timeline = False
 
-    def init(self, topology: networkx.DiGraph = None,
+    def init(self, topology_fn: Callable[[int], networkx.DiGraph] = None,
              is_weighted: bool = False):
         """A function that initializes BlueFog.
 
         Args:
-          topology: A networkx. DiGraph object to decide the topology. If not provided
-            a default power_two_ring structure is used.
-          is_weighted: If set to true, the neighbor ops like (win_sync, neighbor_allreduce) will
+          topology_fn: A callable function that takes size as input and return
+            networkx.DiGraph object to decide the topology. If not provided
+            a default power_two_ring structure is called.
+          is_weighted: If set to true, the neighbor ops like (win_update, neighbor_allreduce) will
             execute the weighted average instead, where the weight is the value used in
             topology matrix (including self).
         """
         self._MPI_LIB_CTYPES.bluefog_init()
-        self.set_topology(topology, is_weighted)
+        if topology_fn:
+            topo = topology_fn(self.size())
+        else:
+            topo = topology_util.PowerTwoRingGraph(self.size())
+        self.set_topology(topo, is_weighted)
         atexit.register(self.shutdown)
 
     def shutdown(self) -> int:
@@ -187,10 +192,10 @@ class BlueFogBasics(object):
         Args:
           Topo: A networkx.DiGraph object to decide the topology. If not provided
             a default power_two_ring structure is used.
-          is_weighted: If set to true, the win_sync and neighbor_allreduce will execute the
+          is_weighted: If set to true, the win_update and neighbor_allreduce will execute the
             weighted average instead, where the weights are the value used in topology matrix
             (including self weight). Note win_get/win_put/win_accumulate do not use this weight
-            since win_sync already uses these weights.
+            since win_update already uses these weights.
 
         Returns:
             A boolean value that whether topology is set correctly or not.
@@ -216,10 +221,10 @@ class BlueFogBasics(object):
             return True
 
         # We remove the self-rank for any cases because MPI graph_comm do not include it.
-        destinations = [r for r in topology.successors(self.rank())
-                        if r != self.rank()]
-        sources = [r for r in topology.predecessors(self.rank())
-                   if r != self.rank()]
+        destinations = sorted([r for r in topology.successors(self.rank())
+                               if r != self.rank()])
+        sources = sorted([r for r in topology.predecessors(self.rank())
+                          if r != self.rank()])
         indegree = len(sources)
         outdegree = len(destinations)
         sources_type = ctypes.c_int * indegree
@@ -265,6 +270,14 @@ class BlueFogBasics(object):
         self._is_topo_weighted = is_weighted
         return True
 
+    def nccl_built(self) -> bool:
+        """Returns True if BlueFog was compiled with NCCL support.
+
+        Returns:
+          A boolean value indicating whether NCCL support was compiled.
+        """
+        return bool(self._MPI_LIB_CTYPES.bluefog_nccl_built())
+
     def timeline_start_activity(self, tensor_name: str, activity_name: str) -> bool:
         """A python interface to call the timeline for StartActivity.
         If you want to use this function, please make sure to turn on the timeline first by
