@@ -26,24 +26,24 @@ namespace common {
 
 class MPIController {
  public:
-  MPIController(TensorQueue& tensor_queue, MPIContext& mpi_ctx)
-      : tensor_queue_(tensor_queue), mpi_ctx_(mpi_ctx) {
+  MPIController(MPIContext& mpi_ctx)
+      : mpi_ctx_(mpi_ctx) {
     BFLOG(DEBUG) << "MPI Controller Initialized.";
   }
   void Initialize();
 
   int GetTypeSize(DataType dtype);
 
-  inline const std::vector<int>& GetRanks() const { return ranks_; };
-  inline int GetRank() const { return rank_; };
-  inline int GetLocalRank() const { return local_rank_; };
-  inline int GetCrossRank() const { return cross_rank_; };
-  inline int GetSize() const { return size_; };
-  inline int GetLocalSize() const { return local_size_; };
-  inline int GetCrossSize() const { return cross_size_; };
-  inline int GetNeighborSize() const { return neighbor_indgree_; };
+  inline const std::vector<int>& GetRanks() const { return mpi_ctx_.ranks_; };
+  inline int GetRank() const { return mpi_ctx_.rank_; };
+  inline int GetLocalRank() const { return mpi_ctx_.local_rank_; };
+  inline int GetCrossRank() const { return mpi_ctx_.cross_rank_; };
+  inline int GetSize() const { return mpi_ctx_.size_; };
+  inline int GetLocalSize() const { return mpi_ctx_.local_size_; };
+  inline int GetCrossSize() const { return mpi_ctx_.cross_size_; };
+  inline int GetNeighborSize() const { return mpi_ctx_.neighbor_indgree_; };
   inline const std::vector<int>& GetLocalCommRanks() const {
-    return local_comm_ranks_;
+    return mpi_ctx_.local_comm_ranks_;
   };
 
   inline bool IsMpiThreadsSupported() const { return mpi_threads_supported_; }
@@ -82,58 +82,29 @@ class MPIController {
   Status WinLock(const std::string& name);
   Status WinUnlock(const std::string& name);
 
-  // Our mutex definition is different from the parallel computation concept.
+  // Our distributed mutex definition is different from the parallel computation concept.
   // For a world size is N application, N mutex is created.
   // Each process associates with one mutex.
+  // Note we create an independent local copy for neighbor process, so there is no conflict
+  // between the writing process from the neighbors (like win_put and win_accumulate). However,
+  // Win_sync (i.e update setup) will read it, which conflicted with other writting process.
   // When WinMutexAcquire is called, we typically lock for all out-neighbors.
-  // For example: if 1 want to win_put value to neighbor 0, and 2 want to
-  // win_accumulate another value to neighbor 0 simultaneously, then only one rank 
-  // can acquire the mutex to 0, i.e. two ops will be serialized.
-  // It is most common used in win_sync (for self) and win_accumulate (for out-neighbor).
-  Status WinMutexAcquire(const std::vector<int>& acquire_ranks);
-  Status WinMutexRelease(const std::vector<int>& release_ranks);
+  Status WinMutexAcquire(const std::vector<int>& acquire_ranks, bool is_sync);
+  Status WinMutexRelease(const std::vector<int>& release_ranks, bool is_sync);
 
  protected:
   // Outside dependencies
-  TensorQueue& tensor_queue_;
-
   MPIContext& mpi_ctx_;
 
   // flag indicating whether MPI multi-threading is supported.
   bool mpi_threads_supported_ = false;
-
-  Status AllocateOutput(TensorTableEntry& entries, int*& recvcounts, Communicator comm_type);
-  void SetDisplacements(const int* recvcounts, int*& displcmnts, Communicator comm_type);
-
- private:
-  int rank_ = 0;
-  int local_rank_ = 0;
-
-  int cross_rank_ = 0;
-  int size_ = 1;
-  int local_size_ = 1;
-  int cross_size_ = 1;
-
-  int neighbor_indgree_ = -1;
-  int neighbor_outdgree_ = -1;
-
-  std::vector<int> neighbor_in_ranks_;
-  std::vector<int> neighbor_out_ranks_;
-
-  // ranks of the bluefog world
-  std::vector<int> ranks_;
-
-  // COMM_WORLD ranks of processes running on this node.
-  std::vector<int> local_comm_ranks_;
-
-  double self_weight_;
-  std::unordered_map<int, double> neighbor_weights_;
 };
 
 class WinMutexGuard {
  public:
   explicit WinMutexGuard(MPIController* mpi_controller,
-                         const std::vector<int>& acquire_ranks);
+                         const std::vector<int>& acquire_ranks,
+                         bool is_sync);
   virtual ~WinMutexGuard();
   WinMutexGuard(const WinMutexGuard&) = delete;
   WinMutexGuard& operator=(const WinMutexGuard&) = delete;
@@ -141,6 +112,7 @@ class WinMutexGuard {
  private:
   MPIController* const mpi_controller_;
   std::vector<int> acquire_ranks_;
+  bool is_sync_;
 };
 
 }  // namespace common
