@@ -31,6 +31,10 @@
 #include "global_state.h"
 #include "logging.h"
 
+#if HAVE_NCCL
+#include "nccl_controller.h"
+#endif
+
 // Bluefog knobs.
 #define BLUEFOG_TIMELINE "BLUEFOG_TIMELINE"
 
@@ -44,6 +48,10 @@ BluefogGlobalState bluefog_global;
 
 MPIContext mpi_context;
 
+#if HAVE_NCCL
+NCCLContext nccl_context;
+#endif
+
 }  // namespace
 
 bool RunLoopOnce(BluefogGlobalState& state);
@@ -55,6 +63,9 @@ void BackgroundThreadLoop(BluefogGlobalState& state) {
   // Initialize controller
   state.controller->Initialize();
 
+  // We use Lazy initialized pattern. nccl_controller will be initialized only
+  // when it necessary. 
+
   // Signal that initialization is completed.
   state.initialization_done = true;
   BFLOG(INFO, bluefog_global.controller->GetRank()) << "Bluefog Initialized";
@@ -88,45 +99,135 @@ void BackgroundThreadLoop(BluefogGlobalState& state) {
     cb(SHUT_DOWN_ERROR);
   }
   mpi_context.Finalize(mpi_ctx_manager);
+
+#if HAVE_NCCL
+  if (nccl_context.is_initialized) {
+    nccl_context.Finalize();
+  }
+#endif
+}
+
+Vendor DetermineController(const TensorTableEntry& entry) {
+  bool nccl_impl_available = false;
+  bool force_mpi = false;
+  bool built_with_nccl = false;
+#if HAVE_NCCL
+  built_with_nccl = true;
+#endif
+  char* by_mpi_env;
+  switch (entry.mpi_ops_type) {
+    case MPIOpsType::ALLREDUCE:
+      by_mpi_env = std::getenv("BLUEFOG_ALLREDUCE_BY_MPI");
+      force_mpi = (by_mpi_env != nullptr) && (*by_mpi_env == '1');
+      nccl_impl_available = true;
+      break;
+    case MPIOpsType::BROADCAST:
+      by_mpi_env = std::getenv("BLUEFOG_BROADCAST_BY_MPI");
+      force_mpi = (by_mpi_env != nullptr) && (*by_mpi_env == '1');
+      nccl_impl_available = true;
+      break;
+    case MPIOpsType::ALLGATHER:
+      by_mpi_env = std::getenv("BLUEFOG_ALLGATHER_BY_MPI");
+      force_mpi = (by_mpi_env != nullptr) && (*by_mpi_env == '1');
+      nccl_impl_available = true;
+      break;
+    case MPIOpsType::NEIGHBOR_ALLGATHER:
+      by_mpi_env = std::getenv("BLUEFOG_NEIGHBOR_ALLGATHER_BY_MPI");
+      force_mpi = (by_mpi_env != nullptr) && (*by_mpi_env == '1');
+      nccl_impl_available = true;
+      break;
+    case MPIOpsType::NEIGHBOR_ALLREDUCE:
+      by_mpi_env = std::getenv("BLUEFOG_NEIGHBOR_ALLREDUCE_BY_MPI");
+      force_mpi = (by_mpi_env != nullptr) && (*by_mpi_env == '1');
+      nccl_impl_available = true;
+      break;
+    default:
+      nccl_impl_available = false;
+  }
+
+  if (!built_with_nccl || !nccl_impl_available || force_mpi) return Vendor::MPI;
+  return entry.device != CPU_DEVICE_ID ? Vendor::NCCL : Vendor::MPI;
 }
 
 bool RunLoopOnce(BluefogGlobalState& state) {
   try {
     auto entry = state.tensor_queue.PopMessagesFromQueue();
+    Vendor controller_vendor = DetermineController(entry);
+#if HAVE_NCCL
+    if(controller_vendor==Vendor::NCCL && !nccl_context.is_initialized) {
+        state.nccl_controller->Initialize();
+        BFLOG(INFO, state.controller->GetRank()) << "NCCL Initialized";
+    }
+#endif
     switch (entry.mpi_ops_type) {
       case MPIOpsType::ALLREDUCE:
         BFLOG(TRACE, bluefog_global.controller->GetRank())
             << "Processing " << entry.tensor_name;
         state.timeline.ActivityStart(entry.tensor_name, MPI_ALLREDUCE);
-        state.controller->Allreduce(entry);
+        if (controller_vendor == Vendor::MPI) {
+          state.controller->Allreduce(entry);
+        }
+#if HAVE_NCCL
+        if (controller_vendor == Vendor::NCCL) {
+          state.nccl_controller->Allreduce(entry);
+        }
+#endif
         state.timeline.ActivityEnd(entry.tensor_name);
         break;
       case MPIOpsType::BROADCAST:
         BFLOG(TRACE, bluefog_global.controller->GetRank())
             << "Processing " << entry.tensor_name;
         state.timeline.ActivityStart(entry.tensor_name, MPI_BROADCAST);
-        state.controller->Broadcast(entry);
+        if (controller_vendor == Vendor::MPI) {
+          state.controller->Broadcast(entry);
+        }
+#if HAVE_NCCL
+        if (controller_vendor == Vendor::NCCL) {
+          state.nccl_controller->Broadcast(entry);
+        }
+#endif
         state.timeline.ActivityEnd(entry.tensor_name);
         break;
       case MPIOpsType::ALLGATHER:
         state.timeline.ActivityStart(entry.tensor_name, MPI_ALLGATHER);
         BFLOG(TRACE, bluefog_global.controller->GetRank())
             << "Processing " << entry.tensor_name;
-        state.controller->Allgather(entry);
+        if (controller_vendor == Vendor::MPI) {
+          state.controller->Allgather(entry);
+        }
+#if HAVE_NCCL
+        if (controller_vendor == Vendor::NCCL) {
+          state.nccl_controller->Allgather(entry);
+        }
+#endif
         state.timeline.ActivityEnd(entry.tensor_name);
         break;
       case MPIOpsType::NEIGHBOR_ALLGATHER:
         state.timeline.ActivityStart(entry.tensor_name, MPI_NEIGHBOR_ALLGATHER);
         BFLOG(TRACE, bluefog_global.controller->GetRank())
             << "Processing " << entry.tensor_name;
-        state.controller->NeighborAllgather(entry);
+        if (controller_vendor == Vendor::MPI) {
+          state.controller->NeighborAllgather(entry);
+        }
+#if HAVE_NCCL
+        if (controller_vendor == Vendor::NCCL) {
+          state.nccl_controller->NeighborAllgather(entry);
+        }
+#endif
         state.timeline.ActivityEnd(entry.tensor_name);
         break;
       case MPIOpsType::NEIGHBOR_ALLREDUCE:
         BFLOG(TRACE, bluefog_global.controller->GetRank())
             << "Processing " << entry.tensor_name;
         state.timeline.ActivityStart(entry.tensor_name, MPI_NEIGHBOR_ALLREDUCE);
-        state.controller->NeighborAllreduce(entry);
+        if (controller_vendor == Vendor::MPI) {
+          state.controller->NeighborAllreduce(entry);
+        }
+#if HAVE_NCCL
+        if (controller_vendor == Vendor::NCCL) {
+          state.nccl_controller->NeighborAllreduce(entry);
+        }
+#endif
         state.timeline.ActivityEnd(entry.tensor_name);
         break;
       case MPIOpsType::BARRIER:
@@ -177,8 +278,11 @@ void InitializeBluefogOnce() {
   mpi_context
       .Enable();  // We always enable mpi since we relied on MPI only now.
   if (!bluefog_global.initialize_flag.test_and_set()) {
-    bluefog_global.controller.reset(
-        new MPIController(bluefog_global.tensor_queue, mpi_context));
+    bluefog_global.controller.reset(new MPIController(mpi_context));
+#if HAVE_NCCL
+    bluefog_global.nccl_controller.reset(
+        new NCCLController(nccl_context, mpi_context));
+#endif
     bluefog_global.initialization_done = false;
     bluefog_global.background_thread =
         std::thread(BackgroundThreadLoop, std::ref(bluefog_global));
@@ -277,8 +381,16 @@ int bluefog_set_topology(int indegree, const int* sources, int outdegree,
         << "Cannot set the topology because there are unfinished MPI ops.";
     return -1;
   }
-  return bluefog_global.controller->SetTopology(indegree, sources, outdegree,
+
+  bool mpi_result = bluefog_global.controller->SetTopology(indegree, sources, outdegree,
                                                 destinations);
+#if HAVE_NCCL && NCCL_MINOR < 7
+  if (mpi_result && nccl_context.is_initialized) {
+    bluefog_global.nccl_controller->DestroyPeerCommunicator();
+    bluefog_global.nccl_controller->InitPeerCommunicator();
+  }
+#endif
+  return mpi_result;
 }
 
 int bluefog_set_topology_with_weights(int indegree, const int* sources,
@@ -332,6 +444,15 @@ int bluefog_timeline(const bool start_activity, const char* tensor_name,
   return 1;
 }
 
+int bluefog_nccl_built() {
+  int result = 0;
+#if HAVE_NCCL
+  result = 1;
+  BFLOG(DEBUG) << "NCCL VERSION: " << NCCL_MAJOR << "." << NCCL_MINOR;
+#endif
+  return result;
+}
+
 }  // extern "C"
 
 Status EnqueueTensorAllreduce(std::shared_ptr<Tensor> tensor,
@@ -587,11 +708,11 @@ Status WindowUnlock(const std::string& name) {
   return status;
 }
 
-Status WindowMutexAcquire(const std::vector<int>& acquire_ranks) {
+Status WindowMutexAcquire(const std::vector<int>& acquire_ranks, bool is_sync) {
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
-  Status status = bluefog_global.controller->WinMutexAcquire(acquire_ranks);
+  Status status = bluefog_global.controller->WinMutexAcquire(acquire_ranks, is_sync);
 
   if (!status.ok()) {
     BFLOG(ERROR) << "Cannot acquire window mutex";
@@ -600,11 +721,11 @@ Status WindowMutexAcquire(const std::vector<int>& acquire_ranks) {
   return status;
 }
 
-Status WindowMutexRelease(const std::vector<int>& release_ranks) {
+Status WindowMutexRelease(const std::vector<int>& release_ranks, bool is_sync) {
   if (bluefog_global.shut_down) {
     return SHUT_DOWN_ERROR;
   }
-  Status status = bluefog_global.controller->WinMutexRelease(release_ranks);
+  Status status = bluefog_global.controller->WinMutexRelease(release_ranks, is_sync);
 
   if (!status.ok()) {
     BFLOG(ERROR) << "Cannot release window mutex"; 
