@@ -92,7 +92,7 @@ int DoAllreduce(::torch::Tensor tensor, ::torch::Tensor output, int average,
     auto bf_output = bf_tensor;
 
     auto enqueue_result = EnqueueTensorAllreduce(
-        bf_tensor, bf_tensor, op_name, device,
+        bf_tensor, bf_tensor, op_name, CPU_DEVICE_ID,
         [handle, average, output, cpu_buffer, device, op_name, tid,
          timeline_ptr](const Status& status) mutable {
           with_device device_guard(device);
@@ -148,7 +148,7 @@ int DoBroadcast(::torch::Tensor tensor, ::torch::Tensor output, int root_rank,
     auto bf_tensor = std::make_shared<TorchTensor>(cpu_buffer);
 
     auto enqueue_result = EnqueueTensorBroadcast(
-        bf_tensor, bf_tensor, root_rank, op_name, device,
+        bf_tensor, bf_tensor, root_rank, op_name, CPU_DEVICE_ID,
         [handle, output, cpu_buffer, device, op_name, tid,
          timeline_ptr](const Status& status) mutable {
           with_device device_guard(device);
@@ -200,7 +200,7 @@ int DoAllgather(::torch::Tensor tensor, ::torch::Tensor output, const std::strin
         std::make_shared<TorchOpContext>(CPU_DEVICE_ID, cpu_output);
 
     auto enqueue_result = EnqueueTensorAllgather(
-        bf_tensor, bf_context, op_name, device,
+        bf_tensor, bf_context, op_name, CPU_DEVICE_ID,
         [handle, cpu_output, device, output, op_name, tid,
          timeline_ptr](const Status& status) mutable {
           with_device device_guard(device);
@@ -250,7 +250,7 @@ int DoNeighborAllgather(::torch::Tensor tensor, ::torch::Tensor output,
         std::make_shared<TorchOpContext>(CPU_DEVICE_ID, cpu_output);
 
     auto enqueue_result = EnqueueTensorNeighborAllgather(
-        bf_tensor, bf_context, op_name, device,
+        bf_tensor, bf_context, op_name, CPU_DEVICE_ID,
         [handle, cpu_output, device, output, op_name, tid,
          timeline_ptr](const Status& status) mutable {
           with_device device_guard(device);
@@ -301,7 +301,7 @@ int DoNeighborAllreduce(::torch::Tensor tensor, ::torch::Tensor output,
         std::make_shared<TorchOpContext>(CPU_DEVICE_ID, cpu_output);
     auto bf_output = std::make_shared<TorchTensor>(cpu_output);
     auto enqueue_result = EnqueueTensorNeighborAllreduce(
-        bf_context, bf_tensor, bf_output, op_name, device,
+        bf_context, bf_tensor, bf_output, op_name, CPU_DEVICE_ID,
         [handle, self_weight, neighbor_weights, avg_computation, cpu_output, tensor,
          output, device, op_name, tid, timeline_ptr](const Status& status) mutable {
           with_device device_guard(device);
@@ -350,7 +350,7 @@ int DoNeighborAllreduce(::torch::Tensor tensor, ::torch::Tensor output,
               }
             }
             output.resize_(shape_vector);
-            output.add_(tensor.mul_(self_weight));
+            output.add_(tensor.mul(self_weight));
           } else {
             auto output_reduced = output.slice(0, 0, first_dim);
             for (int i = 1; i < bluefog_neighbor_size(); i++) {
@@ -418,7 +418,7 @@ int DoNeighborAllreduce(::torch::Tensor tensor, ::torch::Tensor output,
               }
             }
             output.resize_(shape_vector);
-            output.add_(tensor.mul_(self_weight));
+            output.add_(tensor.mul(self_weight));
           } else {
             auto output_reduced = output.slice(0, 0, first_dim);
             for (int i = 1; i < bluefog_neighbor_size(); i++) {
@@ -466,90 +466,90 @@ void AddWinOpsIntoPybind(py::module &);
 
 PYBIND11_MODULE(mpi_lib, m) {
   // allreduce
-  m.def("bluefog_torch_allreduce_async_torch_IntTensor", &DoAllreduce);
-  m.def("bluefog_torch_allreduce_async_torch_LongTensor", &DoAllreduce);
-  m.def("bluefog_torch_allreduce_async_torch_FloatTensor", &DoAllreduce);
-  m.def("bluefog_torch_allreduce_async_torch_DoubleTensor", &DoAllreduce);
+  m.def("bluefog_torch_allreduce_nonblocking_torch_IntTensor", &DoAllreduce);
+  m.def("bluefog_torch_allreduce_nonblocking_torch_LongTensor", &DoAllreduce);
+  m.def("bluefog_torch_allreduce_nonblocking_torch_FloatTensor", &DoAllreduce);
+  m.def("bluefog_torch_allreduce_nonblocking_torch_DoubleTensor", &DoAllreduce);
 #if HAVE_CUDA
-  m.def("bluefog_torch_allreduce_async_torch_cuda_IntTensor", &DoAllreduce);
-  m.def("bluefog_torch_allreduce_async_torch_cuda_LongTensor", &DoAllreduce);
-  m.def("bluefog_torch_allreduce_async_torch_cuda_FloatTensor", &DoAllreduce);
-  m.def("bluefog_torch_allreduce_async_torch_cuda_DoubleTensor", &DoAllreduce);
+  m.def("bluefog_torch_allreduce_nonblocking_torch_cuda_IntTensor", &DoAllreduce);
+  m.def("bluefog_torch_allreduce_nonblocking_torch_cuda_LongTensor", &DoAllreduce);
+  m.def("bluefog_torch_allreduce_nonblocking_torch_cuda_FloatTensor", &DoAllreduce);
+  m.def("bluefog_torch_allreduce_nonblocking_torch_cuda_DoubleTensor", &DoAllreduce);
 #endif
 
   // broadcast
-  m.def("bluefog_torch_broadcast_async_torch_ByteTensor", &DoBroadcast);
-  m.def("bluefog_torch_broadcast_async_torch_CharTensor", &DoBroadcast);
-  m.def("bluefog_torch_broadcast_async_torch_ShortTensor", &DoBroadcast);
-  m.def("bluefog_torch_broadcast_async_torch_IntTensor", &DoBroadcast);
-  m.def("bluefog_torch_broadcast_async_torch_LongTensor", &DoBroadcast);
-  m.def("bluefog_torch_broadcast_async_torch_FloatTensor", &DoBroadcast);
-  m.def("bluefog_torch_broadcast_async_torch_DoubleTensor", &DoBroadcast);
+  m.def("bluefog_torch_broadcast_nonblocking_torch_ByteTensor", &DoBroadcast);
+  m.def("bluefog_torch_broadcast_nonblocking_torch_CharTensor", &DoBroadcast);
+  m.def("bluefog_torch_broadcast_nonblocking_torch_ShortTensor", &DoBroadcast);
+  m.def("bluefog_torch_broadcast_nonblocking_torch_IntTensor", &DoBroadcast);
+  m.def("bluefog_torch_broadcast_nonblocking_torch_LongTensor", &DoBroadcast);
+  m.def("bluefog_torch_broadcast_nonblocking_torch_FloatTensor", &DoBroadcast);
+  m.def("bluefog_torch_broadcast_nonblocking_torch_DoubleTensor", &DoBroadcast);
 #if HAVE_CUDA
-  m.def("bluefog_torch_broadcast_async_torch_cuda_IntTensor", &DoBroadcast);
-  m.def("bluefog_torch_broadcast_async_torch_cuda_LongTensor", &DoBroadcast);
-  m.def("bluefog_torch_broadcast_async_torch_cuda_FloatTensor", &DoBroadcast);
-  m.def("bluefog_torch_broadcast_async_torch_cuda_DoubleTensor", &DoBroadcast);
+  m.def("bluefog_torch_broadcast_nonblocking_torch_cuda_IntTensor", &DoBroadcast);
+  m.def("bluefog_torch_broadcast_nonblocking_torch_cuda_LongTensor", &DoBroadcast);
+  m.def("bluefog_torch_broadcast_nonblocking_torch_cuda_FloatTensor", &DoBroadcast);
+  m.def("bluefog_torch_broadcast_nonblocking_torch_cuda_DoubleTensor", &DoBroadcast);
 #endif
 
   // allgather
-  m.def("bluefog_torch_allgather_async_torch_ByteTensor", &DoAllgather);
-  m.def("bluefog_torch_allgather_async_torch_CharTensor", &DoAllgather);
-  m.def("bluefog_torch_allgather_async_torch_ShortTensor", &DoAllgather);
-  m.def("bluefog_torch_allgather_async_torch_IntTensor", &DoAllgather);
-  m.def("bluefog_torch_allgather_async_torch_LongTensor", &DoAllgather);
-  m.def("bluefog_torch_allgather_async_torch_FloatTensor", &DoAllgather);
-  m.def("bluefog_torch_allgather_async_torch_DoubleTensor", &DoAllgather);
+  m.def("bluefog_torch_allgather_nonblocking_torch_ByteTensor", &DoAllgather);
+  m.def("bluefog_torch_allgather_nonblocking_torch_CharTensor", &DoAllgather);
+  m.def("bluefog_torch_allgather_nonblocking_torch_ShortTensor", &DoAllgather);
+  m.def("bluefog_torch_allgather_nonblocking_torch_IntTensor", &DoAllgather);
+  m.def("bluefog_torch_allgather_nonblocking_torch_LongTensor", &DoAllgather);
+  m.def("bluefog_torch_allgather_nonblocking_torch_FloatTensor", &DoAllgather);
+  m.def("bluefog_torch_allgather_nonblocking_torch_DoubleTensor", &DoAllgather);
 #if HAVE_CUDA
-  m.def("bluefog_torch_allgather_async_torch_cuda_IntTensor", &DoAllgather);
-  m.def("bluefog_torch_allgather_async_torch_cuda_LongTensor", &DoAllgather);
-  m.def("bluefog_torch_allgather_async_torch_cuda_FloatTensor", &DoAllgather);
-  m.def("bluefog_torch_allgather_async_torch_cuda_DoubleTensor", &DoAllgather);
+  m.def("bluefog_torch_allgather_nonblocking_torch_cuda_IntTensor", &DoAllgather);
+  m.def("bluefog_torch_allgather_nonblocking_torch_cuda_LongTensor", &DoAllgather);
+  m.def("bluefog_torch_allgather_nonblocking_torch_cuda_FloatTensor", &DoAllgather);
+  m.def("bluefog_torch_allgather_nonblocking_torch_cuda_DoubleTensor", &DoAllgather);
 #endif
 
   // neighbor_allgather
-  m.def("bluefog_torch_neighbor_allgather_async_torch_ByteTensor",
+  m.def("bluefog_torch_neighbor_allgather_nonblocking_torch_ByteTensor",
         &DoNeighborAllgather);
-  m.def("bluefog_torch_neighbor_allgather_async_torch_CharTensor",
+  m.def("bluefog_torch_neighbor_allgather_nonblocking_torch_CharTensor",
         &DoNeighborAllgather);
-  m.def("bluefog_torch_neighbor_allgather_async_torch_ShortTensor",
+  m.def("bluefog_torch_neighbor_allgather_nonblocking_torch_ShortTensor",
         &DoNeighborAllgather);
-  m.def("bluefog_torch_neighbor_allgather_async_torch_IntTensor",
+  m.def("bluefog_torch_neighbor_allgather_nonblocking_torch_IntTensor",
         &DoNeighborAllgather);
-  m.def("bluefog_torch_neighbor_allgather_async_torch_LongTensor",
+  m.def("bluefog_torch_neighbor_allgather_nonblocking_torch_LongTensor",
         &DoNeighborAllgather);
-  m.def("bluefog_torch_neighbor_allgather_async_torch_FloatTensor",
+  m.def("bluefog_torch_neighbor_allgather_nonblocking_torch_FloatTensor",
         &DoNeighborAllgather);
-  m.def("bluefog_torch_neighbor_allgather_async_torch_DoubleTensor",
+  m.def("bluefog_torch_neighbor_allgather_nonblocking_torch_DoubleTensor",
         &DoNeighborAllgather);
 #if HAVE_CUDA
-  m.def("bluefog_torch_neighbor_allgather_async_torch_cuda_IntTensor",
+  m.def("bluefog_torch_neighbor_allgather_nonblocking_torch_cuda_IntTensor",
         &DoNeighborAllgather);
-  m.def("bluefog_torch_neighbor_allgather_async_torch_cuda_LongTensor",
+  m.def("bluefog_torch_neighbor_allgather_nonblocking_torch_cuda_LongTensor",
         &DoNeighborAllgather);
-  m.def("bluefog_torch_neighbor_allgather_async_torch_cuda_FloatTensor",
+  m.def("bluefog_torch_neighbor_allgather_nonblocking_torch_cuda_FloatTensor",
         &DoNeighborAllgather);
-  m.def("bluefog_torch_neighbor_allgather_async_torch_cuda_DoubleTensor",
+  m.def("bluefog_torch_neighbor_allgather_nonblocking_torch_cuda_DoubleTensor",
         &DoNeighborAllgather);
 #endif
 
   // neighbor_allreduce
-  m.def("bluefog_torch_neighbor_allreduce_async_torch_IntTensor",
+  m.def("bluefog_torch_neighbor_allreduce_nonblocking_torch_IntTensor",
         &DoNeighborAllreduce);
-  m.def("bluefog_torch_neighbor_allreduce_async_torch_LongTensor",
+  m.def("bluefog_torch_neighbor_allreduce_nonblocking_torch_LongTensor",
         &DoNeighborAllreduce);
-  m.def("bluefog_torch_neighbor_allreduce_async_torch_FloatTensor",
+  m.def("bluefog_torch_neighbor_allreduce_nonblocking_torch_FloatTensor",
         &DoNeighborAllreduce);
-  m.def("bluefog_torch_neighbor_allreduce_async_torch_DoubleTensor",
+  m.def("bluefog_torch_neighbor_allreduce_nonblocking_torch_DoubleTensor",
         &DoNeighborAllreduce);
 #if HAVE_CUDA
-  m.def("bluefog_torch_neighbor_allreduce_async_torch_cuda_IntTensor",
+  m.def("bluefog_torch_neighbor_allreduce_nonblocking_torch_cuda_IntTensor",
         &DoNeighborAllreduce);
-  m.def("bluefog_torch_neighbor_allreduce_async_torch_cuda_LongTensor",
+  m.def("bluefog_torch_neighbor_allreduce_nonblocking_torch_cuda_LongTensor",
         &DoNeighborAllreduce);
-  m.def("bluefog_torch_neighbor_allreduce_async_torch_cuda_FloatTensor",
+  m.def("bluefog_torch_neighbor_allreduce_nonblocking_torch_cuda_FloatTensor",
         &DoNeighborAllreduce);
-  m.def("bluefog_torch_neighbor_allreduce_async_torch_cuda_DoubleTensor",
+  m.def("bluefog_torch_neighbor_allreduce_nonblocking_torch_cuda_DoubleTensor",
         &DoNeighborAllreduce);
 #endif
 
