@@ -39,6 +39,7 @@ in_neighbor_ranks = _basics.in_neighbor_ranks
 out_neighbor_ranks = _basics.out_neighbor_ranks
 mpi_threads_supported = _basics.mpi_threads_supported
 unified_mpi_window_model_supported = _basics.unified_mpi_window_model_supported
+nccl_built = _basics.nccl_built
 
 timeline_context = _basics.timeline_context
 timeline_start_activity = _basics.timeline_start_activity
@@ -67,10 +68,10 @@ def _check_function(function_factory, tensor, *args):
 
 
 def _allreduce_function_factory(tensor):
-    return 'bluefog_torch_allreduce_async_' + tensor.type().replace('.', '_')
+    return 'bluefog_torch_allreduce_nonblocking_' + tensor.type().replace('.', '_')
 
 
-def _allreduce_async(tensor, output, average, name):
+def _allreduce_nonblocking(tensor, output, average, name):
     function = _check_function(_allreduce_function_factory, tensor)
     if average:
         assert isinstance(tensor, (torch.FloatTensor, torch.DoubleTensor,
@@ -103,13 +104,13 @@ def allreduce(tensor: torch.Tensor, average: bool = True, name: str = None) -> t
         A tensor of the same shape and type as `tensor`, averaged or summed across all
         processes.
     """
-    handle = allreduce_async(tensor, average, name)
+    handle = allreduce_nonblocking(tensor, average, name)
     return synchronize(handle)
 
 
-def allreduce_async(tensor: torch.Tensor, average: bool = True, name: str = None) -> int:
+def allreduce_nonblocking(tensor: torch.Tensor, average: bool = True, name: str = None) -> int:
     """
-    A function that performs asynchronous averaging or summation of the input tensor
+    A function that performs nonblocking averaging or summation of the input tensor
     over all the Bluefog processes. The input tensor is not modified.
 
     The reduction operation is keyed by the name. If name is not provided, an incremented
@@ -128,14 +129,14 @@ def allreduce_async(tensor: torch.Tensor, average: bool = True, name: str = None
         `synchronize()`.
     """
     output = tensor.new(tensor.shape)
-    return _allreduce_async(tensor, output, average, name)
+    return _allreduce_nonblocking(tensor, output, average, name)
 
 
 def _broadcast_function_factory(tensor):
-    return 'bluefog_torch_broadcast_async_' + tensor.type().replace('.', '_')
+    return 'bluefog_torch_broadcast_nonblocking_' + tensor.type().replace('.', '_')
 
 
-def _broadcast_async(tensor, output, root_rank, name):
+def _broadcast_nonblocking(tensor, output, root_rank, name):
     function = _check_function(_broadcast_function_factory, tensor)
     handle = getattr(mpi_lib, function)(tensor, output, root_rank,
                                         name.encode() if name is not None else "")
@@ -166,13 +167,13 @@ def broadcast(tensor: torch.Tensor, root_rank: int, name: str = None) -> torch.T
         A tensor of the same shape and type as `tensor`, with the value broadcasted
         from root rank.
     """
-    handle = broadcast_async(tensor, root_rank, name)
+    handle = broadcast_nonblocking(tensor, root_rank, name)
     return synchronize(handle)
 
 
-def broadcast_async(tensor: torch.Tensor, root_rank: int, name: str = None) -> int:
+def broadcast_nonblocking(tensor: torch.Tensor, root_rank: int, name: str = None) -> int:
     """
-    A function that asynchronously broadcasts the input tensor on root rank to the same
+    A function that nonblockingly broadcasts the input tensor on root rank to the same
     input tensor on all other Bluefog processes. The input tensor is not modified.
 
     The broadcast operation is keyed by the name. If name is not provided, an incremented
@@ -190,7 +191,7 @@ def broadcast_async(tensor: torch.Tensor, root_rank: int, name: str = None) -> i
         `synchronize()`.
     """
     output = tensor.new(tensor.shape)
-    return _broadcast_async(tensor, output, root_rank, name)
+    return _broadcast_nonblocking(tensor, output, root_rank, name)
 
 
 def broadcast_(tensor, root_rank, name=None) -> torch.Tensor:
@@ -212,13 +213,13 @@ def broadcast_(tensor, root_rank, name=None) -> torch.Tensor:
         A tensor of the same shape and type as `tensor`, with the value broadcasted
         from root rank.
     """
-    handle = broadcast_async_(tensor, root_rank, name)
+    handle = broadcast_nonblocking_(tensor, root_rank, name)
     return synchronize(handle)
 
 
-def broadcast_async_(tensor, root_rank, name=None) -> int:
+def broadcast_nonblocking_(tensor, root_rank, name=None) -> int:
     """
-    A function that asynchronously broadcasts the input tensor on root rank to the same
+    A function that nonblockingly broadcasts the input tensor on root rank to the same
     input tensor on all other Bluefog processes. The operation is performed in-place.
 
     The broadcast operation is keyed by the name. If name is not provided, an incremented
@@ -235,14 +236,14 @@ def broadcast_async_(tensor, root_rank, name=None) -> int:
         A handle to the broadcast operation that can be used with `poll()` or
         `synchronize()`.
     """
-    return _broadcast_async(tensor, tensor, root_rank, name)
+    return _broadcast_nonblocking(tensor, tensor, root_rank, name)
 
 
 def _allgather_function_factory(tensor):
-    return 'bluefog_torch_allgather_async_' + tensor.type().replace('.', '_')
+    return 'bluefog_torch_allgather_nonblocking_' + tensor.type().replace('.', '_')
 
 
-def _allgather_async(tensor, output, name):
+def _allgather_nonblocking(tensor, output, name):
     function = _check_function(_allgather_function_factory, tensor)
     handle = getattr(mpi_lib, function)(tensor, output,
                                         name.encode() if name is not None else "")
@@ -268,13 +269,13 @@ def allgather(tensor: torch.Tensor, name: str = None) -> torch.Tensor:
         the first dimension, which may be greater and is the sum of all first
         dimensions of the tensors in different Bluefog processes.
     """
-    handle = allgather_async(tensor, name)
+    handle = allgather_nonblocking(tensor, name)
     return synchronize(handle)
 
 
-def allgather_async(tensor: torch.Tensor, name: str = None) -> int:
+def allgather_nonblocking(tensor: torch.Tensor, name: str = None) -> int:
     """
-    A function that asynchronously concatenates the input tensor with the same input
+    A function that nonblockingly concatenates the input tensor with the same input
     tensor on all other Bluefog processes. The input tensor is not modified.
 
     The concatenation is done on the first dimension, so the input tensors on the
@@ -289,14 +290,14 @@ def allgather_async(tensor: torch.Tensor, name: str = None) -> int:
         `synchronize()`.
     """
     output = tensor.new()  # real size will be allocated later.
-    return _allgather_async(tensor, output, name)
+    return _allgather_nonblocking(tensor, output, name)
 
 
 def _neighbor_allgather_function_factory(tensor):
-    return 'bluefog_torch_neighbor_allgather_async_' + tensor.type().replace('.', '_')
+    return 'bluefog_torch_neighbor_allgather_nonblocking_' + tensor.type().replace('.', '_')
 
 
-def _neighbor_allgather_async(tensor, output, name):
+def _neighbor_allgather_nonblocking(tensor, output, name):
     function = _check_function(_neighbor_allgather_function_factory, tensor)
     handle = getattr(mpi_lib, function)(tensor, output,
                                         name.encode() if name is not None else "")
@@ -322,13 +323,13 @@ def neighbor_allgather(tensor: torch.Tensor, name: str = None) -> torch.Tensor:
         the first dimension, which may be greater and is the sum of all first
         dimensions of the tensors in neighbor Bluefog processes.
     """
-    handle = neighbor_allgather_async(tensor, name)
+    handle = neighbor_allgather_nonblocking(tensor, name)
     return synchronize(handle)
 
 
-def neighbor_allgather_async(tensor: torch.Tensor, name: str = None) -> int:
+def neighbor_allgather_nonblocking(tensor: torch.Tensor, name: str = None) -> int:
     """
-    A function that asynchronously concatenates the input tensor with the same input
+    A function that nonblockingly concatenates the input tensor with the same input
     tensor on all neighbor Bluefog processes (Not include self).
     The input tensor is not modified.
 
@@ -344,14 +345,14 @@ def neighbor_allgather_async(tensor: torch.Tensor, name: str = None) -> int:
         `synchronize()`.
     """
     output = tensor.new()  # real size will be allocated later.
-    return _neighbor_allgather_async(tensor, output, name)
+    return _neighbor_allgather_nonblocking(tensor, output, name)
 
 
 def _neighbor_allreduce_function_factory(tensor):
-    return 'bluefog_torch_neighbor_allreduce_async_' + tensor.type().replace('.', '_')
+    return 'bluefog_torch_neighbor_allreduce_nonblocking_' + tensor.type().replace('.', '_')
 
 
-def _neighbor_allreduce_async(tensor, output, self_weight, neighbor_weights, name):
+def _neighbor_allreduce_nonblocking(tensor, output, self_weight, neighbor_weights, name):
     function = _check_function(_neighbor_allreduce_function_factory, tensor)
     if self_weight is None and neighbor_weights is None:
         if is_topo_weighted():
@@ -361,14 +362,15 @@ def _neighbor_allreduce_async(tensor, output, self_weight, neighbor_weights, nam
         else:
             weight = 1.0/(len(in_neighbor_ranks())+1)
             self_weight = weight
-            neighbor_weights = {r:weight for r in in_neighbor_ranks()}
+            neighbor_weights = {r: weight for r in in_neighbor_ranks()}
             avg_computation = False
     elif self_weight is not None and neighbor_weights is not None:
         if not isinstance(neighbor_weights, dict):
             raise ValueError("Argument neighbor_weights has to be a dictionary map from the "
                              "(in-)neighbor rank to the weights.")
         if not isinstance(self_weight, float):
-            raise ValueError("Argument self_weight has to be a float for self rank.")
+            raise ValueError(
+                "Argument self_weight has to be a float for self rank.")
         if not set(neighbor_weights.keys()).issubset(set(in_neighbor_ranks())):
             raise ValueError("The key of weights should only contain the ranks that belong to "
                              " in-neighbors and self rank.")
@@ -416,15 +418,17 @@ def neighbor_allreduce(tensor: torch.Tensor,
        (self_weight is not None and neighbor_weights is None):
         raise ValueError("Arguments self_weight and neighbor_weights have to be presented at "
                          "the same time")
-    handle = neighbor_allreduce_async(tensor, self_weight, neighbor_weights, name)
+    handle = neighbor_allreduce_nonblocking(
+        tensor, self_weight, neighbor_weights, name)
     return synchronize(handle)
 
 
-def neighbor_allreduce_async(tensor: torch.Tensor,
-                             self_weight: float = None, neighbor_weights: Dict[int, float] = None,
-                             name: str = None) -> int:
+def neighbor_allreduce_nonblocking(tensor: torch.Tensor,
+                                   self_weight: float = None,
+                                   neighbor_weights: Dict[int, float] = None,
+                                   name: str = None) -> int:
     """
-    A function that asynchronously performs weighted averaging of the input tensor over the
+    A function that nonblockingly performs weighted averaging of the input tensor over the
     negihbors and itself in the Bluefog processes. The default behavior is (uniformly) average.
 
     The input tensor is not modified.
@@ -456,18 +460,18 @@ def neighbor_allreduce_async(tensor: torch.Tensor,
         raise ValueError("Arguments self_weight and neighbor_weights have to be presented at "
                          "the same time")
     output = tensor.new(tensor.shape)
-    return _neighbor_allreduce_async(tensor, output, self_weight, neighbor_weights, name)
+    return _neighbor_allreduce_nonblocking(tensor, output, self_weight, neighbor_weights, name)
 
 
 def poll(handle: int) -> bool:
     """
     Polls an allreduce, allgather or broadcast handle to determine whether underlying
-    asynchronous operation has completed. After `poll()` returns `True`, `synchronize()`
+    nonblocking operation has completed. After `poll()` returns `True`, `synchronize()`
     will return without blocking.
 
     Arguments:
-        handle: A handle returned by an allreduce, allgather or broadcast asynchronous
-                operation.
+        handle: A handle returned by an allreduce, allgather, broadcast, neighbor_allgather,
+        and neighbro_allreduce nonblocking operation.
 
     Returns:
         A flag indicating whether the operation has completed.
@@ -477,11 +481,11 @@ def poll(handle: int) -> bool:
 
 def synchronize(handle: int) -> torch.Tensor:
     """
-    Synchronizes an asynchronous allreduce, allgather or broadcast operation until
+    Synchronizes an nonblocking allreduce, allgather or broadcast operation until
     it's completed. Returns the result of the operation.
 
     Args:
-        handle: A handle returned by an allreduce, allgather or broadcast asynchronous
+        handle: A handle returned by an allreduce, allgather or broadcast nonblocking
                 operation.
 
     Returns:
@@ -527,7 +531,7 @@ def win_create(tensor: torch.Tensor, name: str, zero_init: bool = False) -> bool
         bool: Indicate the creation succeed or not.
 
     Note: The window with same name across different bluefog processes should associate
-    the tensor with same shape. Otherwise, the rest win_ops like win_sync, win_put will
+    the tensor with same shape. Otherwise, the rest win_ops like win_update, win_put will
     encounter unrecoverable memory segmentation fault.
     """
     function = _check_function(_win_create_function_factory, tensor)
@@ -555,15 +559,16 @@ def win_free(name: str = None) -> bool:
     return getattr(mpi_lib, 'bluefog_torch_win_free')(name)
 
 
-def _win_sync_function_factory(tensor):
+def _win_update_function_factory(tensor):
     return 'bluefog_torch_win_sync_' + tensor.type().replace('.', '_')
 
 
-def win_sync_then_collect(name: str) -> torch.Tensor:
+def win_update_then_collect(name: str, require_mutex=True) -> torch.Tensor:
     """ A utility function to sync the neighbor buffers then accumulate all
     neighbor buffers' tensors into self tensor and clear the buffer.
     It is equivalent to
-    >>> win_sync(name, self_weight=1.0, neighbor_weights={neighbor: 1.0}, reset=True)
+    >>> win_update(name, self_weight=1.0, neighbor_weights={neighbor: 1.0}, reset=True,
+                   require_mutex=require_mutex)
 
     Args:
         name: The unique name to associate the window object.
@@ -572,15 +577,16 @@ def win_sync_then_collect(name: str) -> torch.Tensor:
         torch.Tensor: The average tensor of all neighbors' cooresponding tensors.
     """
     neighbor_weights = {r: 1.0 for r in in_neighbor_ranks()}
-    return win_sync(name, 1.0, neighbor_weights, reset=True)
+    return win_update(name, 1.0, neighbor_weights, reset=True, require_mutex=require_mutex)
 
 
-def win_sync(name: str,
-             self_weight: float = None, neighbor_weights: Dict[int, float] = None,
-             reset: bool = False, clone: bool = False) -> torch.Tensor:
+def win_update(name: str,
+               self_weight: float = None, neighbor_weights: Dict[int, float] = None,
+               reset: bool = False, clone: bool = False,
+               require_mutex: bool = False) -> torch.Tensor:
     """Locally synchronized the window objects and returned the reduced neighbor tensor.
     Note the returned tensor is the same tensor used in win_create and in-place modification
-    is happened.
+    is happened. During the update, a mutex for local variable is acquired.
 
     Args:
         name: The unique name to associate the window object.
@@ -595,8 +601,10 @@ def win_sync(name: str,
             neighbor_weights will be reset to zero.
             The reset is always happened after the weights computation.
             If neighbor_weights is not presented and reset is True, all the neighbor will be reset.
-        clone: If set up to be true, the win_sync result will return a new tensor instead of
+        clone: If set up to be true, the win_update result will return a new tensor instead of
             in-place change.
+        require_mutex: If set to be true, the window mutex associated with local process will be
+            acquired.
 
     Returns:
         torch.Tensor: The average tensor of all neighbors' cooresponding tensors.
@@ -605,14 +613,12 @@ def win_sync(name: str,
     change with the iterations. If static weight need, then setting the weights through the
     bf.set_topology(.., is_weighted=True) is a better choice.
 
-    Note2: If reset is True, mutex for self is acquired.
-
-    Note3: self_weight and neighbor_weights must be presented at the same time.
+    Note2: self_weight and neighbor_weights must be presented at the same time.
     """
     tensor = _win_map[name]
     if clone:
         tensor = tensor.clone()
-    function = _check_function(_win_sync_function_factory, tensor)
+    function = _check_function(_win_update_function_factory, tensor)
 
     if neighbor_weights is not None and self_weight is not None:
         # Pre-condition check for weights dictionary.
@@ -620,7 +626,8 @@ def win_sync(name: str,
             raise ValueError("Argument neighbor_weights has to be a dictionary map from the "
                              "(in-)neighbor rank to the weights.")
         if not isinstance(self_weight, float):
-            raise ValueError("Argument self_weight has to be a float for self rank.")
+            raise ValueError(
+                "Argument self_weight has to be a float for self rank.")
         if not set(neighbor_weights.keys()).issubset(set(in_neighbor_ranks())):
             raise ValueError("The key of weights should only contain the ranks that belong to "
                              " in-neighbors and self rank.")
@@ -634,15 +641,15 @@ def win_sync(name: str,
         else:
             weight = 1.0/(len(in_neighbor_ranks())+1)
             self_weight = weight
-            neighbor_weights = {r:weight for r in in_neighbor_ranks()}
+            neighbor_weights = {r: weight for r in in_neighbor_ranks()}
             avg_computation = False
     else:
         raise ValueError("Arguments self_weight and neighbor_weights have to be presented at "
                          "the same time")
 
     if not getattr(mpi_lib, function)(tensor, name, self_weight, neighbor_weights,
-                                      reset, avg_computation):
-        raise RuntimeError("Cannot apply win_sync on " + name)
+                                      reset, avg_computation, require_mutex):
+        raise RuntimeError("Cannot apply win_update on " + name)
     return tensor
 
 
@@ -659,9 +666,9 @@ def _win_put_function_factory(tensor):
     return 'bluefog_torch_win_put_' + tensor.type().replace('.', '_')
 
 
-def win_put_async(tensor: torch.Tensor, name: str,
-                  dst_weights: Dict[int, float] = None,
-                  require_mutex: bool = False) -> int:
+def win_put_nonblocking(tensor: torch.Tensor, name: str,
+                        dst_weights: Dict[int, float] = None,
+                        require_mutex: bool = False) -> int:
     """ Passively put the tensor into neighbor's shared window memory.
     This is a non-blocking function, which will return without waiting the
     win_put operation is really finished.
@@ -688,7 +695,8 @@ def win_put_async(tensor: torch.Tensor, name: str,
         raise ValueError(
             "The key of dst_weights should only containranks that "
             " belong to out-neighbors (self-rank is not allowed).")
-    handle = getattr(mpi_lib, function)(tensor, name, dst_weights, require_mutex)
+    handle = getattr(mpi_lib, function)(
+        tensor, name, dst_weights, require_mutex)
     _win_handle_map[handle] = name
     return handle
 
@@ -714,15 +722,15 @@ def win_put(tensor: torch.Tensor, name: str,
     Returns:
         A bool value to indicate the put succeeded or not.
     """
-    handle = win_put_async(tensor, name, dst_weights, require_mutex)
+    handle = win_put_nonblocking(tensor, name, dst_weights, require_mutex)
     return win_wait(handle)
 
 
-def win_get_async(name: str, src_weights: Dict[int, float] = None,
-                  require_mutex: bool = False) -> int:
+def win_get_nonblocking(name: str, src_weights: Dict[int, float] = None,
+                        require_mutex: bool = False) -> int:
     """ Passively get the tensor(s) from neighbors' shared window memory into
     local shared memory, which cannot be accessed in python directly.
-    The win_sync function is responsible for fetching that memeory.
+    The win_update function is responsible for fetching that memeory.
     This is a non-blocking function, which will return without waiting the
     win_get operation is really finished.
 
@@ -756,7 +764,7 @@ def win_get(name: str, src_weights: Dict[int, float] = None,
             require_mutex: bool = False) -> bool:
     """ Passively get the tensor(s) from neighbors' shared window memory into
     local shared memory, which cannot be accessed in python directly.
-    The win_sync function is responsible for fetching that memeory.
+    The win_update function is responsible for fetching that memeory.
     This is a blocking function, which will return until win_get operation
     is finished.
 
@@ -774,10 +782,9 @@ def win_get(name: str, src_weights: Dict[int, float] = None,
             acquired.
 
     Returns:
-        A tensor of the same shape and type as `tensor`, averaged or summed across src_ranks
-        processes (or all neighbor processes).
+        A bool value to indicate the get succeeded or not.
     """
-    handle = win_get_async(name, src_weights, require_mutex)
+    handle = win_get_nonblocking(name, src_weights, require_mutex)
     return win_wait(handle)
 
 
@@ -785,9 +792,9 @@ def _win_accumulate_function_factory(tensor):
     return 'bluefog_torch_win_accumulate_' + tensor.type().replace('.', '_')
 
 
-def win_accumulate_async(tensor: torch.Tensor, name: str,
-                         dst_weights: Dict[int, float] = None,
-                         require_mutex: bool = False) -> bool:
+def win_accumulate_nonblocking(tensor: torch.Tensor, name: str,
+                               dst_weights: Dict[int, float] = None,
+                               require_mutex: bool = False) -> bool:
     """ Passively accmulate the tensor into neighbor's shared window memory.
     Only SUM ops is supported now.
     This is a non-blocking function, which will return without waiting the
@@ -841,10 +848,10 @@ def win_accumulate(tensor: torch.Tensor, name: str,
             acquired.
 
     Returns:
-        A handle to the win_accmulate operation that can be used with `win_poll()` or
-        `win_wait()`.
+        A bool value to indicate the accumulate succeeded or not.
     """
-    handle = win_accumulate_async(tensor, name, dst_weights, require_mutex)
+    handle = win_accumulate_nonblocking(
+        tensor, name, dst_weights, require_mutex)
     return win_wait(handle)
 
 
@@ -895,31 +902,33 @@ def _win_unlock(name: str):
 
 
 @contextmanager
-def win_mutex(ranks: List[int] = None):
+def win_mutex(ranks: List[int] = None, exclusive=True):
     """ A win object implemented mutex context manager. Note, there are N distributed
     mutex over N corresponding processes.
 
     Args:
         ranks: The mutex associated with the ranks will be acquired.
             If not presented, out_neighbor ranks will be used.
+        exclusive: If not exclusive, multiple processes can hold the mutex for the ranks
+            but it cannot hold with one exclusive one.
 
     Example:
         >>> bf.win_create(tensor, name)
         >>> with win_mutex():
-                tensor = bf.win_sync_then_collect(name)
+                tensor = bf.win_update_then_collect(name)
         >>> win_put(tensor, name)
     """
     _ranks = out_neighbor_ranks() if ranks is None else ranks
-    _win_mutex_acquire(_ranks)
+    _win_mutex_acquire(_ranks, exclusive)
     try:
         yield
     finally:
-        _win_mutex_release(_ranks)
+        _win_mutex_release(_ranks, exclusive)
 
 
-def _win_mutex_acquire(ranks):
-    mpi_lib.bluefog_torch_win_mutex_acquire(ranks)
+def _win_mutex_acquire(ranks, exclusive):
+    mpi_lib.bluefog_torch_win_mutex_acquire(ranks, exclusive)
 
 
-def _win_mutex_release(ranks):
-    mpi_lib.bluefog_torch_win_mutex_release(ranks)
+def _win_mutex_release(ranks, exclusive):
+    mpi_lib.bluefog_torch_win_mutex_release(ranks, exclusive)
