@@ -158,6 +158,16 @@ bool WinTorchStorageManager::GetStorageByNameRank(
   return true;
 }
 
+bool WinTorchStorageManager::SetSelfStorageByName(
+    const std::string& name, const ::torch::Tensor& tensor) {
+  auto it = self_tensor_map_.find(name);
+  if (it == self_tensor_map_.end()) {
+    return false;
+  }
+  it->second->GetUnderlyingTensor().copy_(tensor);
+  return true;
+}
+
 bool WinTorchStorageManager::SumWithNeighbor(const std::string& name,
                                              ::torch::Tensor local_tensor) {
   auto it = tensors_map_.find(name);
@@ -267,19 +277,20 @@ int ResetNeighborTensor(const std::string& name,
 int DoWinSync(::torch::Tensor tensor, const std::string& name,
               double self_weight,
               const std::unordered_map<int, double>& neighbor_weights,
-              bool reset, bool internal_avg) {
+              bool reset, bool internal_avg, bool require_mutex) {
   ThrowIfError(common::CheckInitialized());
 
   Timeline* timeline_ptr;
   Status timeline_status = GetBluefogTimeline(timeline_ptr);
   timeline_ptr->ActivityStart(name, "WIN_SYNC_COMPUTE_AVERAGE");
 
-  // We need to lock self avoid updating and win_put/win_accumulate happen at simultaneous time.
+  // We need to lock self avoid updating and win_put/win_accumulate happen at
+  // simultaneous time.
   const std::vector<int> self_rank = {common::bluefog_rank()};
-  if (reset && !neighbor_weights.empty()) common::WindowMutexAcquire(self_rank);
+  if (require_mutex) common::WindowMutexAcquire(self_rank, /*is_sync=*/true);
 
   int device = CPU_DEVICE_ID;
-  if(!win_storage_manager.GetDeviceByName(name, &device)) {
+  if (!win_storage_manager.GetDeviceByName(name, &device)) {
     BFLOG(ERROR) << "Cannot get device of win " << name;
     return 0;
   }
@@ -288,35 +299,44 @@ int DoWinSync(::torch::Tensor tensor, const std::string& name,
 
   ::torch::Tensor cpu_buffer = tensor;
   if (WIN_ON_CPU && tensor.device().is_cuda()) {
-    cpu_buffer = tensor.to(::torch::Device(::torch::kCPU), /*non_blocking=*/false);
+    cpu_buffer =
+        tensor.to(::torch::Device(::torch::kCPU), /*non_blocking=*/false);
   }
 
-  // internal_avg specifies the detailed flow for weighted reduction operation for the neighbors
-  // which may lead to efficiency and precision difference.
-  // but when internal_avg is false, the results are only correct when all weights are
-  // 1/(neighbor size+1).
+  // internal_avg specifies the detailed flow for weighted reduction operation
+  // for the neighbors which may lead to efficiency and precision difference.
+  // but when internal_avg is false, the results are only correct when all
+  // weights are 1/(neighbor size+1).
   if (internal_avg) {
     // Weighted averaging with neighbors' tensors happens in-place.
-    if (!win_storage_manager.AvgWithNeighbor(name, cpu_buffer, self_weight, neighbor_weights))
+    if (!win_storage_manager.AvgWithNeighbor(name, cpu_buffer, self_weight,
+                                             neighbor_weights)) {
+      if (require_mutex) common::WindowMutexRelease(self_rank, /*is_sync=*/true);
       return 0;
+    }
   } else {
     // Sum over neighbors' tensors happens in-place.
     if (!win_storage_manager.SumWithNeighbor(name, cpu_buffer)) {
+      if (require_mutex) common::WindowMutexRelease(self_rank, /*is_sync=*/true);
       return 0;
     }
     // +1 here because in neighbor degree doesn't include self rank.
-    double neighbor_size = neighbor_weights.size()+1.0;
+    double neighbor_size = neighbor_weights.size() + 1.0;
     cpu_buffer.div_(neighbor_size);
   }
 
-  if (reset && !ResetNeighborTensor(name, neighbor_weights)) return 0;
-
-  if (reset && !neighbor_weights.empty()) common::WindowMutexRelease(self_rank);
+  if (reset && !ResetNeighborTensor(name, neighbor_weights)) {
+    if (require_mutex) common::WindowMutexRelease(self_rank, /*is_sync=*/true);
+    return 0;
+  }
+  if (require_mutex) common::WindowMutexRelease(self_rank, /*is_sync=*/true);
 
   if (WIN_ON_CPU && tensor.device().is_cuda()) {
     auto device = GetDeviceID(tensor);
     with_device device_guard(device);
     tensor.copy_(cpu_buffer);
+    // TODO(ybc) When pull_get is used in GPU, we need to copy twice here. Optimizer it.
+    // win_storage_manager.SetSelfStorageByName(name, tensor);
   }
   timeline_ptr->ActivityEnd(name);  // WIN_SYNC_COMPUTE_AVERAGE
 
@@ -368,6 +388,7 @@ int DoWinPut(::torch::Tensor tensor, const std::string& name,
     ::torch::Tensor cpu_buffer =
         tensor.to(::torch::Device(::torch::kCPU), /*non_blocking=*/false);
     bf_tensor = std::make_shared<TorchTensor>(cpu_buffer);
+    device = CPU_DEVICE_ID;
   } else {
     bf_tensor = std::make_shared<TorchTensor>(tensor);
   }
@@ -402,6 +423,7 @@ int DoWinAccumulate(::torch::Tensor tensor, const std::string& name,
     ::torch::Tensor cpu_buffer =
         tensor.to(::torch::Device(::torch::kCPU), /*non_blocking=*/false);
     bf_tensor = std::make_shared<TorchTensor>(cpu_buffer);
+    device = CPU_DEVICE_ID;
   } else {
     bf_tensor = std::make_shared<TorchTensor>(tensor);
   }
@@ -416,7 +438,6 @@ int DoWinAccumulate(::torch::Tensor tensor, const std::string& name,
       });
 
   ThrowIfError(enqueue_result);
-  timeline_ptr->ActivityEnd(name);  // ENQUEUE
   return handle;
 }
 
@@ -486,15 +507,15 @@ void DoWinUnlock(const std::string& name) {
   ThrowIfError(status);
 }
 
-void DoWinMutexAcquire(const std::vector<int>& ranks) {
+void DoWinMutexAcquire(const std::vector<int>& ranks, bool exclusive) {
   ThrowIfError(common::CheckInitialized());
-  Status status = common::WindowMutexAcquire(ranks);
+  Status status = common::WindowMutexAcquire(ranks, /*is_sync=*/exclusive);
   ThrowIfError(status);
 }
 
-void DoWinMutexRelease(const std::vector<int>& ranks) {
+void DoWinMutexRelease(const std::vector<int>& ranks, bool exclusive) {
   ThrowIfError(common::CheckInitialized());
-  Status status = common::WindowMutexRelease(ranks);
+  Status status = common::WindowMutexRelease(ranks, /*is_sync=*/exclusive);
   ThrowIfError(status);
 }
 
